from __future__ import annotations

from warnings import catch_warnings, filterwarnings

from travertino.properties.aliased import Condition, aliased_property
from travertino.properties.choices import Choices
from travertino.properties.shorthand import directional_property
from travertino.properties.validated import list_property, validated_property
from travertino.style import BaseStyle

from ..utils import apply_dataclass, mock_apply

VALUE1 = "value1"
VALUE2 = "value2"
VALUE3 = "value3"
VALUES = [VALUE1, VALUE2, VALUE3, None]


@mock_apply
@apply_dataclass
class Style(BaseStyle):
    # Some properties with explicit initial values
    explicit_const: str | int = validated_property(
        *VALUES, integer=True, initial=VALUE1
    )
    explicit_value: str | int = validated_property(*VALUES, integer=True, initial=0)
    explicit_none: str | int | None = validated_property(
        *VALUES, integer=True, initial=None
    )

    # A property with an implicit default value.
    # This usually means the default is platform specific.
    implicit: str | int | None = validated_property(
        VALUE1, VALUE2, VALUE3, integer=True
    )

    # A set of directional properties
    thing: tuple[str | int] | str | int = directional_property("thing{}")
    thing_top: str | int = validated_property(*VALUES, integer=True, initial=0)
    thing_right: str | int = validated_property(*VALUES, integer=True, initial=0)
    thing_bottom: str | int = validated_property(*VALUES, integer=True, initial=0)
    thing_left: str | int = validated_property(*VALUES, integer=True, initial=0)

    # Nothing below here needs to be tested in deprecated API:
    list_prop: list[str] = list_property(*VALUES, integer=True, initial=(VALUE2,))

    # A variety of aliases to other properties
    plain_alias: str | int = aliased_property(source="explicit_const")
    plain_alias_deprecated: str | int = aliased_property(
        source="explicit_const", deprecated=True
    )
    directional_alias: tuple[str | int] | str | int = aliased_property(source="thing")
    directional_alias_deprecated: tuple[str | int] | str | int = aliased_property(
        source="thing", deprecated=True
    )
    conditional_alias: str | int = aliased_property(
        source={
            Condition(thing_top=0): "explicit_const",
            Condition(thing_top=10, list_prop=[VALUE1, VALUE2]): "explicit_value",
            Condition(thing_top=10, list_prop=[VALUE1]): "explicit_none",
        }
    )
    conditional_alias_deprecated: str | int = aliased_property(
        source={
            Condition(thing_top=0): "explicit_const",
            Condition(thing_top=10, list_prop=[VALUE1, VALUE2]): "explicit_value",
            Condition(thing_top=10, list_prop=[VALUE1]): "explicit_none",
        },
        deprecated=True,
    )


VALUE_CHOICES = Choices(*VALUES, integer=True)

with catch_warnings():
    filterwarnings("ignore", category=DeprecationWarning)

    @mock_apply
    class DeprecatedStyle(BaseStyle):
        pass

    # Some properties with explicit initial values
    DeprecatedStyle.validated_property(
        "explicit_const", choices=VALUE_CHOICES, initial=VALUE1
    )
    DeprecatedStyle.validated_property(
        "explicit_value", choices=VALUE_CHOICES, initial=0
    )
    DeprecatedStyle.validated_property(
        "explicit_none", choices=VALUE_CHOICES, initial=None
    )

    # A property with an implicit default value.
    # This usually means the default is platform specific.
    DeprecatedStyle.validated_property(
        "implicit", choices=Choices(VALUE1, VALUE2, VALUE3, integer=True)
    )

    # A set of directional properties
    DeprecatedStyle.validated_property("thing_top", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.validated_property("thing_right", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.validated_property("thing_bottom", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.validated_property("thing_left", choices=VALUE_CHOICES, initial=0)
    DeprecatedStyle.directional_property("thing%s")



from unittest.mock import call

import pytest

from .style_classes import DeprecatedStyle, Style


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_directional_property(StyleClass):
    style = StyleClass()

    # Default value is 0
    assert style.thing == (0, 0, 0, 0)
    assert style.thing_top == 0
    assert style.thing_right == 0
    assert style.thing_bottom == 0
    assert style.thing_left == 0
    assert "thing" not in style
    style.apply.assert_not_called()

    # Set a value in one axis
    style.thing_top = 10

    assert style.thing == (10, 0, 0, 0)
    assert style.thing_top == 10
    assert style.thing_right == 0
    assert style.thing_bottom == 0
    assert style.thing_left == 0
    assert "thing" in style
    style.apply.assert_called_once_with("thing_top")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with a single item
    style.thing = (10,)

    assert style.thing == (10, 10, 10, 10)
    assert style.thing_top == 10
    assert style.thing_right == 10
    assert style.thing_bottom == 10
    assert style.thing_left == 10
    assert "thing" in style
    style.apply.assert_has_calls(
        [
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with a single item
    style.thing = 30

    assert style.thing == (30, 30, 30, 30)
    assert style.thing_top == 30
    assert style.thing_right == 30
    assert style.thing_bottom == 30
    assert style.thing_left == 30
    assert "thing" in style
    style.apply.assert_has_calls(
        [
            call("thing_top"),
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with 2 values
    style.thing = (10, 20)

    assert style.thing == (10, 20, 10, 20)
    assert style.thing_top == 10
    assert style.thing_right == 20
    assert style.thing_bottom == 10
    assert style.thing_left == 20
    assert "thing" in style
    style.apply.assert_has_calls(
        [
            call("thing_top"),
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with 3 values
    style.thing = (10, 20, 30)

    assert style.thing == (10, 20, 30, 20)
    assert style.thing_top == 10
    assert style.thing_right == 20
    assert style.thing_bottom == 30
    assert style.thing_left == 20
    assert "thing" in style
    style.apply.assert_called_once_with("thing_bottom")

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set a value directly with 4 values
    style.thing = (10, 20, 30, 40)

    assert style.thing == (10, 20, 30, 40)
    assert style.thing_top == 10
    assert style.thing_right == 20
    assert style.thing_bottom == 30
    assert style.thing_left == 40
    assert "thing" in style
    style.apply.assert_called_once_with("thing_left")

    # Set a value directly with an invalid number of values
    with pytest.raises(ValueError):
        style.thing = ()

    with pytest.raises(ValueError):
        style.thing = (10, 20, 30, 40, 50)

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear a value on one axis
    del style.thing_top

    assert style.thing == (0, 20, 30, 40)
    assert style.thing_top == 0
    assert style.thing_right == 20
    assert style.thing_bottom == 30
    assert style.thing_left == 40
    assert "thing" in style
    style.apply.assert_called_once_with("thing_top")

    # Restore the top thing
    style.thing_top = 10

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear a value directly
    del style.thing

    assert style.thing == (0, 0, 0, 0)
    assert style.thing_top == 0
    assert style.thing_right == 0
    assert style.thing_bottom == 0
    assert style.thing_left == 0
    assert "thing" not in style
    style.apply.assert_has_calls(
        [
            call("thing_right"),
            call("thing_bottom"),
            call("thing_left"),
        ]
    )



from unittest.mock import Mock, call

from .style_classes import VALUE2, VALUE3, Style


def test_batched_apply():
    """With applicator, apply should batch calls to internal _apply."""
    style = Style()
    style._applicator = Mock()
    style.apply.reset_mock()
    style._apply.reset_mock()

    style.update(explicit_const=VALUE2, implicit=VALUE3)
    # Apply is called once for each property during update, but these calls are stored
    # and batched into one combined call to _apply.
    assert style.apply.mock_calls == [call("explicit_const"), call("implicit")]
    style._apply.assert_called_once_with({"explicit_const", "implicit"})

    style.apply.reset_mock()
    style._apply.reset_mock()


def test_batched_apply_no_names():
    """Batching doesn't do anything if no names were batched."""
    style = Style()
    style._applicator = Mock()
    style._apply.reset_mock()

    with style.batch_apply():
        pass

    style._apply.assert_not_called()


def test_batched_apply_nested():
    """Nesting batch_apply() does nothing; _apply is only called when all are exited."""
    style = Style()
    style._applicator = Mock()
    style._apply.reset_mock()

    with style.batch_apply():
        with style.batch_apply():
            style.update(explicit_const=VALUE2, implicit=VALUE3)

        style._apply.assert_not_called()

    style._apply.assert_called_once_with({"explicit_const", "implicit"})


def test_batched_apply_directional():
    """Assigning or deleting a directional property batches to _apply."""
    style = Style()
    style._applicator = Mock()
    style._apply.reset_mock()

    style.thing = 10
    style._apply.assert_called_once_with(
        {"thing_top", "thing_right", "thing_bottom", "thing_left"}
    )
    style._apply.reset_mock()

    del style.thing
    style._apply.assert_called_once_with(
        {"thing_top", "thing_right", "thing_bottom", "thing_left"}
    )
    style._apply.reset_mock()

    style.thing = (15, 15, 15, 15)
    style._apply.assert_called_once_with(
        {"thing_top", "thing_right", "thing_bottom", "thing_left"}
    )
    style._apply.reset_mock()

    style.thing = (0, 15, 15, 15)
    style._apply.assert_called_once_with({"thing_top"})
    style._apply.reset_mock()

    del style.thing
    style._apply.assert_called_once_with({"thing_right", "thing_bottom", "thing_left"})



from __future__ import annotations

from .constants import *  # noqa: F403


def _clamp(value, upper=1):
    return min(upper, max(0, value))


def _clamp_hex(value):
    return round(_clamp(value, upper=255))


class Color:
    "A base class for all colorspace representations."

    def __eq__(self, other):
        try:
            c1 = self.rgba
            c2 = other.rgba

            return c1.r == c2.r and c1.g == c2.g and c1.b == c2.b and c1.a == c2.a
        except AttributeError:
            return False

    @classmethod
    def _validate_between(cls, content_name, value, min_value, max_value):
        if value < min_value or value > max_value:
            raise ValueError(
                f"{content_name} value should be between {min_value}-{max_value}. "
                f"Got {value}"
            )

    @classmethod
    def _validate_partial(cls, content_name, value):
        cls._validate_between(content_name, value, 0, 1)

    @classmethod
    def _validate_alpha(cls, value):
        cls._validate_partial("alpha", value)

    def blend_over(
        self,
        back_color: Color,
    ) -> rgba:
        """Performs the "over" straight alpha blending operation, compositing
        the front color over the back color.

        **Straight alpha blending** is not the same as
        **Premultiplied alpha blending**, see:
        https://en.wikipedia.org/wiki/Alpha_compositing#Straight_versus_premultiplied

        :param back_color: The background color.

        :returns: The blended color.
        """
        # The blending operation implemented here is the "over" operation and
        # replicates CSS's rgba mechanism. For the formulae used here, see:
        # https://en.wikipedia.org/wiki/Alpha_compositing#Description

        # Convert the input colors to rgba in order to do the calculation.
        front_color = self.rgba
        back_color = back_color.rgba

        if front_color.a == 1:
            # If the front color is fully opaque then the result will be the same as
            # front color.
            return front_color

        blended_alpha = _clamp(front_color.a + ((1 - front_color.a) * back_color.a))

        if blended_alpha == 0:
            # Don't further blend the color, to prevent divide by 0.
            return rgba(0, 0, 0, 0)
        else:
            bands = {}
            for band in "rgb":
                front = getattr(front_color, band)
                back = getattr(back_color, band)
                bands[band] = _clamp_hex(
                    (
                        (front * front_color.a)
                        + (back * back_color.a * (1 - front_color.a))
                    )
                    / blended_alpha
                )

            return rgba(**bands, a=blended_alpha)

    def unblend_over(self, back_color: Color, front_color_alpha: float) -> rgba:
        """Performs the reverse of the "over" straight alpha blending operation,
        returning the front color.

        Note: Unblending of blended colors might produce front color with slightly
        imprecise component values compared to the original front color. This is
        due to the loss of some amount of precision during the calculation and
        conversion process between the different color formats. For example,
        unblending of a hsla blended color might might produce a slightly imprecise
        original front color, since the alpha blending and unblending is calculated
        after conversion to rgba values.

        :param back_color: The background color.
        :param front_color_alpha: The original alpha value of the front color,
            within the range of (0, 1].

        :raises ValueError: If the value of :any:`front_color_alpha` is not within
            the range of (0, 1]. The value cannot be 0, since the blended color produced
            will be equal to the back color, and all information related to the front
            color will be lost.

        :returns: The original front color.
        """
        # The blending operation implemented here is the reverse of the "over"
        # operation and replicates CSS's rgba mechanism. The formula used here
        # are derived from the "over" straight alpha blending operation formula,
        # see: https://en.wikipedia.org/wiki/Alpha_compositing#Description

        blended_color = self.rgba
        back_color = back_color.rgba
        if not 0 < front_color_alpha <= 1:
            raise ValueError(
                "The value of front_color_alpha must be within the range of (0, 1]."
            )
        else:
            bands = {}
            for band in "rgb":
                blended = getattr(blended_color, band)
                back = getattr(back_color, band)
                bands[band] = _clamp_hex(
                    (
                        (blended * blended_color.a)
                        - (back * back_color.a * (1 - front_color_alpha))
                    )
                    / front_color_alpha
                )

            return rgba(**bands, a=front_color_alpha)


class rgba(Color):
    "A representation of an RGBA color."

    def __init__(self, r, g, b, a):
        self._validate_rgb("red", r)
        self._validate_rgb("green", g)
        self._validate_rgb("blue", b)
        self._validate_alpha(a)
        self.r = r
        self.g = g
        self.b = b
        self.a = a

    def __hash__(self):
        return hash(("RGBA-color", self.r, self.g, self.b, self.a))

    def __repr__(self):
        return f"rgba({self.r}, {self.g}, {self.b}, {self.a})"

    @classmethod
    def _validate_rgb(cls, content_name, value):
        cls._validate_between(content_name, value, 0, 255)

    @property
    def rgba(self):
        # Explicitly return a rgba value, to ensure that classes which inherit the
        # parent rgba class, actually return the rgba() value, instead of the child
        # class. For example, without this fix, doing rgb(20, 20, 20).rgba will
        # return rgb(20, 20, 20), instead of rgba(20, 20, 20, 1).
        return rgba(self.r, self.g, self.b, self.a)

    @property
    def rgb(self):
        return rgb(self.r, self.g, self.b)

    @property
    def hsla(self) -> hsla:
        # Formula used here is from: https://en.wikipedia.org/wiki/HSL_and_HSV#From_RGB
        r_prime = self.r / 255
        g_prime = self.g / 255
        b_prime = self.b / 255

        max_component = max(r_prime, g_prime, b_prime)
        min_component = min(r_prime, g_prime, b_prime)
        value = max_component
        chroma = max_component - min_component

        lightness = (max_component + min_component) / 2

        if chroma == 0:
            hue = 0
        elif value == r_prime:
            hue = 60 * (((g_prime - b_prime) / chroma) % 6)
        elif value == g_prime:
            hue = 60 * (((b_prime - r_prime) / chroma) + 2)
        else:  # value == b_prime:
            hue = 60 * (((r_prime - g_prime) / chroma) + 4)

        if lightness in {0, 1}:
            saturation = 0
        else:
            saturation = chroma / (1 - abs((2 * value) - chroma - 1))

        return hsla(
            hue % 360,  # [0,360)
            _clamp(saturation),  # [0,1]
            _clamp(lightness),  # [0,1]
            self.a,  # [0,1]
        )

    @property
    def hsl(self):
        return self.hsla.hsl


class rgb(rgba):
    "A representation of an RGB color"

    def __init__(self, r, g, b):
        super().__init__(r, g, b, 1.0)

    def __repr__(self):
        return f"rgb({self.r}, {self.g}, {self.b})"


class hsla(Color):
    "A representation of an HSLA color."

    def __init__(self, h, s, l, a=1.0):  # noqa: E741
        self._validate_between("hue", h, 0, 360)
        self._validate_partial("saturation", s)
        self._validate_partial("lightness", l)
        self._validate_alpha(a)
        self.h = h
        self.s = s
        self.l = l
        self.a = a

    def __hash__(self):
        return hash(("HSLA-color", self.h, self.s, self.l, self.a))

    def __repr__(self):
        return f"hsla({self.h}, {self.s}, {self.l}, {self.a})"

    @property
    def hsla(self):
        # Explicitly return a hsla value, to ensure that classes which inherit the
        # parent hsla class, actually return the hsla() value, instead of the child
        # class. For example, without this fix, doing hsl(210, 1, 0.5).hsla will
        # return hsl(210, 1, 0.5), instead of hsla(210, 1, 0.5, 1).
        return hsla(self.h, self.s, self.l, self.a)

    @property
    def hsl(self):
        return hsl(self.h, self.s, self.l)

    @property
    def rgba(self):
        c = (1.0 - abs(2.0 * self.l - 1.0)) * self.s
        h = self.h / 60.0
        x = c * (1.0 - abs(h % 2 - 1.0))
        m = self.l - 0.5 * c

        if h < 1.0:
            r, g, b = c + m, x + m, m
        elif h < 2.0:
            r, g, b = x + m, c + m, m
        elif h < 3.0:
            r, g, b = m, c + m, x + m
        elif h < 4.0:
            r, g, b = m, x + m, c + m
        elif h < 5.0:
            r, g, b = x + m, m, c + m
        else:
            r, g, b = c + m, m, x + m

        return rgba(
            round(r * 0xFF),
            round(g * 0xFF),
            round(b * 0xFF),
            self.a,
        )

    @property
    def rgb(self):
        return self.rgba.rgb


class hsl(hsla):
    "A representation of an HSL color"

    def __init__(self, h, s, l):  # noqa: E741
        super().__init__(h, s, l, 1.0)

    def __repr__(self):
        return f"hsl({self.h}, {self.s}, {self.l})"


def color(value):
    """Parse a color from a value.

    Accepts:
    * rgb() instances
    * hsl() instances
    * '#rgb'
    * '#rgba'
    * '#rrggbb'
    * '#rrggbbaa'
    * '#RGB'
    * '#RGBA'
    * '#RRGGBB'
    * '#RRGGBBAA'
    * 'rgb(0, 0, 0)'
    * 'rgba(0, 0, 0, 0.0)'
    * 'hsl(0, 0%, 0%)'
    * 'hsla(0, 0%, 0%, 0.0)'
    * A named color
    """

    if isinstance(value, Color):
        return value

    elif isinstance(value, str):
        if value[0] == "#":
            if len(value) == 4:
                return rgb(
                    r=int(value[1] + value[1], 16),
                    g=int(value[2] + value[2], 16),
                    b=int(value[3] + value[3], 16),
                )
            elif len(value) == 5:
                return rgba(
                    r=int(value[1] + value[1], 16),
                    g=int(value[2] + value[2], 16),
                    b=int(value[3] + value[3], 16),
                    a=int(value[4] + value[4], 16) / 0xFF,
                )
            elif len(value) == 7:
                return rgb(
                    r=int(value[1:3], 16),
                    g=int(value[3:5], 16),
                    b=int(value[5:7], 16),
                )
            elif len(value) == 9:
                return rgba(
                    r=int(value[1:3], 16),
                    g=int(value[3:5], 16),
                    b=int(value[5:7], 16),
                    a=int(value[7:9], 16) / 0xFF,
                )
        elif value.startswith("rgba"):
            try:
                values = value[5:-1].split(",")
                if len(values) == 4:
                    return rgba(
                        int(values[0]),
                        int(values[1]),
                        int(values[2]),
                        float(
                            values[3],
                        ),
                    )
            except ValueError:
                pass
        elif value.startswith("rgb"):
            try:
                values = value[4:-1].split(",")
                if len(values) == 3:
                    return rgb(
                        int(values[0]),
                        int(values[1]),
                        int(values[2]),
                    )
            except ValueError:
                pass

        elif value.startswith("hsla"):
            try:
                values = value[5:-1].split(",")
                if len(values) == 4:
                    return hsla(
                        int(values[0]),
                        int(values[1].strip().rstrip("%")) / 100.0,
                        int(values[2].strip().rstrip("%")) / 100.0,
                        float(values[3]),
                    )
            except ValueError:
                pass

        elif value.startswith("hsl"):
            try:
                values = value[4:-1].split(",")
                if len(values) == 3:
                    return hsl(
                        int(values[0]),
                        int(values[1].strip().rstrip("%")) / 100.0,
                        int(values[2].strip().rstrip("%")) / 100.0,
                    )
            except ValueError:
                pass
        else:
            try:
                return NAMED_COLOR[value.lower()]
            except KeyError:
                pass

    raise ValueError(f"Unknown color {value}")


NAMED_COLOR = {
    ALICEBLUE: rgb(0xF0, 0xF8, 0xFF),  # noqa: F405
    ANTIQUEWHITE: rgb(0xFA, 0xEB, 0xD7),  # noqa: F405
    AQUA: rgb(0x00, 0xFF, 0xFF),  # noqa: F405
    AQUAMARINE: rgb(0x7F, 0xFF, 0xD4),  # noqa: F405
    AZURE: rgb(0xF0, 0xFF, 0xFF),  # noqa: F405
    BEIGE: rgb(0xF5, 0xF5, 0xDC),  # noqa: F405
    BISQUE: rgb(0xFF, 0xE4, 0xC4),  # noqa: F405
    BLACK: rgb(0x00, 0x00, 0x00),  # noqa: F405
    BLANCHEDALMOND: rgb(0xFF, 0xEB, 0xCD),  # noqa: F405
    BLUE: rgb(0x00, 0x00, 0xFF),  # noqa: F405
    BLUEVIOLET: rgb(0x8A, 0x2B, 0xE2),  # noqa: F405
    BROWN: rgb(0xA5, 0x2A, 0x2A),  # noqa: F405
    BURLYWOOD: rgb(0xDE, 0xB8, 0x87),  # noqa: F405
    CADETBLUE: rgb(0x5F, 0x9E, 0xA0),  # noqa: F405
    CHARTREUSE: rgb(0x7F, 0xFF, 0x00),  # noqa: F405
    CHOCOLATE: rgb(0xD2, 0x69, 0x1E),  # noqa: F405
    CORAL: rgb(0xFF, 0x7F, 0x50),  # noqa: F405
    CORNFLOWERBLUE: rgb(0x64, 0x95, 0xED),  # noqa: F405
    CORNSILK: rgb(0xFF, 0xF8, 0xDC),  # noqa: F405
    CRIMSON: rgb(0xDC, 0x14, 0x3C),  # noqa: F405
    CYAN: rgb(0x00, 0xFF, 0xFF),  # noqa: F405
    DARKBLUE: rgb(0x00, 0x00, 0x8B),  # noqa: F405
    DARKCYAN: rgb(0x00, 0x8B, 0x8B),  # noqa: F405
    DARKGOLDENROD: rgb(0xB8, 0x86, 0x0B),  # noqa: F405
    DARKGRAY: rgb(0xA9, 0xA9, 0xA9),  # noqa: F405
    DARKGREY: rgb(0xA9, 0xA9, 0xA9),  # noqa: F405
    DARKGREEN: rgb(0x00, 0x64, 0x00),  # noqa: F405
    DARKKHAKI: rgb(0xBD, 0xB7, 0x6B),  # noqa: F405
    DARKMAGENTA: rgb(0x8B, 0x00, 0x8B),  # noqa: F405
    DARKOLIVEGREEN: rgb(0x55, 0x6B, 0x2F),  # noqa: F405
    DARKORANGE: rgb(0xFF, 0x8C, 0x00),  # noqa: F405
    DARKORCHID: rgb(0x99, 0x32, 0xCC),  # noqa: F405
    DARKRED: rgb(0x8B, 0x00, 0x00),  # noqa: F405
    DARKSALMON: rgb(0xE9, 0x96, 0x7A),  # noqa: F405
    DARKSEAGREEN: rgb(0x8F, 0xBC, 0x8F),  # noqa: F405
    DARKSLATEBLUE: rgb(0x48, 0x3D, 0x8B),  # noqa: F405
    DARKSLATEGRAY: rgb(0x2F, 0x4F, 0x4F),  # noqa: F405
    DARKSLATEGREY: rgb(0x2F, 0x4F, 0x4F),  # noqa: F405
    DARKTURQUOISE: rgb(0x00, 0xCE, 0xD1),  # noqa: F405
    DARKVIOLET: rgb(0x94, 0x00, 0xD3),  # noqa: F405
    DEEPPINK: rgb(0xFF, 0x14, 0x93),  # noqa: F405
    DEEPSKYBLUE: rgb(0x00, 0xBF, 0xFF),  # noqa: F405
    DIMGRAY: rgb(0x69, 0x69, 0x69),  # noqa: F405
    DIMGREY: rgb(0x69, 0x69, 0x69),  # noqa: F405
    DODGERBLUE: rgb(0x1E, 0x90, 0xFF),  # noqa: F405
    FIREBRICK: rgb(0xB2, 0x22, 0x22),  # noqa: F405
    FLORALWHITE: rgb(0xFF, 0xFA, 0xF0),  # noqa: F405
    FORESTGREEN: rgb(0x22, 0x8B, 0x22),  # noqa: F405
    FUCHSIA: rgb(0xFF, 0x00, 0xFF),  # noqa: F405
    GAINSBORO: rgb(0xDC, 0xDC, 0xDC),  # noqa: F405
    GHOSTWHITE: rgb(0xF8, 0xF8, 0xFF),  # noqa: F405
    GOLD: rgb(0xFF, 0xD7, 0x00),  # noqa: F405
    GOLDENROD: rgb(0xDA, 0xA5, 0x20),  # noqa: F405
    GRAY: rgb(0x80, 0x80, 0x80),  # noqa: F405
    GREY: rgb(0x80, 0x80, 0x80),  # noqa: F405
    GREEN: rgb(0x00, 0x80, 0x00),  # noqa: F405
    GREENYELLOW: rgb(0xAD, 0xFF, 0x2F),  # noqa: F405
    HONEYDEW: rgb(0xF0, 0xFF, 0xF0),  # noqa: F405
    HOTPINK: rgb(0xFF, 0x69, 0xB4),  # noqa: F405
    INDIANRED: rgb(0xCD, 0x5C, 0x5C),  # noqa: F405
    INDIGO: rgb(0x4B, 0x00, 0x82),  # noqa: F405
    IVORY: rgb(0xFF, 0xFF, 0xF0),  # noqa: F405
    KHAKI: rgb(0xF0, 0xE6, 0x8C),  # noqa: F405
    LAVENDER: rgb(0xE6, 0xE6, 0xFA),  # noqa: F405
    LAVENDERBLUSH: rgb(0xFF, 0xF0, 0xF5),  # noqa: F405
    LAWNGREEN: rgb(0x7C, 0xFC, 0x00),  # noqa: F405
    LEMONCHIFFON: rgb(0xFF, 0xFA, 0xCD),  # noqa: F405
    LIGHTBLUE: rgb(0xAD, 0xD8, 0xE6),  # noqa: F405
    LIGHTCORAL: rgb(0xF0, 0x80, 0x80),  # noqa: F405
    LIGHTCYAN: rgb(0xE0, 0xFF, 0xFF),  # noqa: F405
    LIGHTGOLDENRODYELLOW: rgb(0xFA, 0xFA, 0xD2),  # noqa: F405
    LIGHTGRAY: rgb(0xD3, 0xD3, 0xD3),  # noqa: F405
    LIGHTGREY: rgb(0xD3, 0xD3, 0xD3),  # noqa: F405
    LIGHTGREEN: rgb(0x90, 0xEE, 0x90),  # noqa: F405
    LIGHTPINK: rgb(0xFF, 0xB6, 0xC1),  # noqa: F405
    LIGHTSALMON: rgb(0xFF, 0xA0, 0x7A),  # noqa: F405
    LIGHTSEAGREEN: rgb(0x20, 0xB2, 0xAA),  # noqa: F405
    LIGHTSKYBLUE: rgb(0x87, 0xCE, 0xFA),  # noqa: F405
    LIGHTSLATEGRAY: rgb(0x77, 0x88, 0x99),  # noqa: F405
    LIGHTSLATEGREY: rgb(0x77, 0x88, 0x99),  # noqa: F405
    LIGHTSTEELBLUE: rgb(0xB0, 0xC4, 0xDE),  # noqa: F405
    LIGHTYELLOW: rgb(0xFF, 0xFF, 0xE0),  # noqa: F405
    LIME: rgb(0x00, 0xFF, 0x00),  # noqa: F405
    LIMEGREEN: rgb(0x32, 0xCD, 0x32),  # noqa: F405
    LINEN: rgb(0xFA, 0xF0, 0xE6),  # noqa: F405
    MAGENTA: rgb(0xFF, 0x00, 0xFF),  # noqa: F405
    MAROON: rgb(0x80, 0x00, 0x00),  # noqa: F405
    MEDIUMAQUAMARINE: rgb(0x66, 0xCD, 0xAA),  # noqa: F405
    MEDIUMBLUE: rgb(0x00, 0x00, 0xCD),  # noqa: F405
    MEDIUMORCHID: rgb(0xBA, 0x55, 0xD3),  # noqa: F405
    MEDIUMPURPLE: rgb(0x93, 0x70, 0xDB),  # noqa: F405
    MEDIUMSEAGREEN: rgb(0x3C, 0xB3, 0x71),  # noqa: F405
    MEDIUMSLATEBLUE: rgb(0x7B, 0x68, 0xEE),  # noqa: F405
    MEDIUMSPRINGGREEN: rgb(0x00, 0xFA, 0x9A),  # noqa: F405
    MEDIUMTURQUOISE: rgb(0x48, 0xD1, 0xCC),  # noqa: F405
    MEDIUMVIOLETRED: rgb(0xC7, 0x15, 0x85),  # noqa: F405
    MIDNIGHTBLUE: rgb(0x19, 0x19, 0x70),  # noqa: F405
    MINTCREAM: rgb(0xF5, 0xFF, 0xFA),  # noqa: F405
    MISTYROSE: rgb(0xFF, 0xE4, 0xE1),  # noqa: F405
    MOCCASIN: rgb(0xFF, 0xE4, 0xB5),  # noqa: F405
    NAVAJOWHITE: rgb(0xFF, 0xDE, 0xAD),  # noqa: F405
    NAVY: rgb(0x00, 0x00, 0x80),  # noqa: F405
    OLDLACE: rgb(0xFD, 0xF5, 0xE6),  # noqa: F405
    OLIVE: rgb(0x80, 0x80, 0x00),  # noqa: F405
    OLIVEDRAB: rgb(0x6B, 0x8E, 0x23),  # noqa: F405
    ORANGE: rgb(0xFF, 0xA5, 0x00),  # noqa: F405
    ORANGERED: rgb(0xFF, 0x45, 0x00),  # noqa: F405
    ORCHID: rgb(0xDA, 0x70, 0xD6),  # noqa: F405
    PALEGOLDENROD: rgb(0xEE, 0xE8, 0xAA),  # noqa: F405
    PALEGREEN: rgb(0x98, 0xFB, 0x98),  # noqa: F405
    PALETURQUOISE: rgb(0xAF, 0xEE, 0xEE),  # noqa: F405
    PALEVIOLETRED: rgb(0xDB, 0x70, 0x93),  # noqa: F405
    PAPAYAWHIP: rgb(0xFF, 0xEF, 0xD5),  # noqa: F405
    PEACHPUFF: rgb(0xFF, 0xDA, 0xB9),  # noqa: F405
    PERU: rgb(0xCD, 0x85, 0x3F),  # noqa: F405
    PINK: rgb(0xFF, 0xC0, 0xCB),  # noqa: F405
    PLUM: rgb(0xDD, 0xA0, 0xDD),  # noqa: F405
    POWDERBLUE: rgb(0xB0, 0xE0, 0xE6),  # noqa: F405
    PURPLE: rgb(0x80, 0x00, 0x80),  # noqa: F405
    REBECCAPURPLE: rgb(0x66, 0x33, 0x99),  # noqa: F405
    RED: rgb(0xFF, 0x00, 0x00),  # noqa: F405
    ROSYBROWN: rgb(0xBC, 0x8F, 0x8F),  # noqa: F405
    ROYALBLUE: rgb(0x41, 0x69, 0xE1),  # noqa: F405
    SADDLEBROWN: rgb(0x8B, 0x45, 0x13),  # noqa: F405
    SALMON: rgb(0xFA, 0x80, 0x72),  # noqa: F405
    SANDYBROWN: rgb(0xF4, 0xA4, 0x60),  # noqa: F405
    SEAGREEN: rgb(0x2E, 0x8B, 0x57),  # noqa: F405
    SEASHELL: rgb(0xFF, 0xF5, 0xEE),  # noqa: F405
    SIENNA: rgb(0xA0, 0x52, 0x2D),  # noqa: F405
    SILVER: rgb(0xC0, 0xC0, 0xC0),  # noqa: F405
    SKYBLUE: rgb(0x87, 0xCE, 0xEB),  # noqa: F405
    SLATEBLUE: rgb(0x6A, 0x5A, 0xCD),  # noqa: F405
    SLATEGRAY: rgb(0x70, 0x80, 0x90),  # noqa: F405
    SLATEGREY: rgb(0x70, 0x80, 0x90),  # noqa: F405
    SNOW: rgb(0xFF, 0xFA, 0xFA),  # noqa: F405
    SPRINGGREEN: rgb(0x00, 0xFF, 0x7F),  # noqa: F405
    STEELBLUE: rgb(0x46, 0x82, 0xB4),  # noqa: F405
    TAN: rgb(0xD2, 0xB4, 0x8C),  # noqa: F405
    TEAL: rgb(0x00, 0x80, 0x80),  # noqa: F405
    THISTLE: rgb(0xD8, 0xBF, 0xD8),  # noqa: F405
    TOMATO: rgb(0xFF, 0x63, 0x47),  # noqa: F405
    TURQUOISE: rgb(0x40, 0xE0, 0xD0),  # noqa: F405
    VIOLET: rgb(0xEE, 0x82, 0xEE),  # noqa: F405
    WHEAT: rgb(0xF5, 0xDE, 0xB3),  # noqa: F405
    WHITE: rgb(0xFF, 0xFF, 0xFF),  # noqa: F405
    WHITESMOKE: rgb(0xF5, 0xF5, 0xF5),  # noqa: F405
    YELLOW: rgb(0xFF, 0xFF, 0x00),  # noqa: F405
    YELLOWGREEN: rgb(0x9A, 0xCD, 0x32),  # noqa: F405
}


__all__ = [
    "Color",
    "rgba",
    "rgb",
    "hsla",
    "hsl",
    "color",
    "NAMED_COLOR",
    "TRANSPARENT",  # noqa: F405
] + [name.upper() for name in NAMED_COLOR]



from unittest.mock import Mock

import pytest

from travertino.properties.choices import Choices

from ..utils import apply_dataclass, mock_apply
from .style_classes import (
    VALUE1,
    VALUE2,
    VALUE3,
    VALUES,
    BaseStyle,
    Style,
)

# Note: Other tests are parametrized with DeprecatedStyle; these are only the tests
# that *only* test deprecated behavior.


@mock_apply
@apply_dataclass
class MockedApplyStyle(BaseStyle):
    pass


def test_deprecated_copy():
    style = MockedApplyStyle()

    with pytest.warns(DeprecationWarning):
        style_copy = style.copy(applicator=object())

    style_copy.apply.assert_called_once_with()


def test_deprecated_class_methods():
    """Toga < 0.5.0 can still define style properties with the old API."""

    class OldStyle(BaseStyle):
        pass

    with pytest.warns(DeprecationWarning):
        OldStyle.validated_property(
            "implicit", Choices(*VALUES, integer=True), initial=VALUE1
        )

    with pytest.warns(DeprecationWarning):
        OldStyle.directional_property("thing%s")


def test_deprecated_reapply():
    """Reapply() is deprecated (but still calls apply()."""
    style = Style()
    with pytest.warns(DeprecationWarning):
        style.reapply()

    style.apply.assert_called_once_with()


def test_deprecated_import():
    """Toga < 0.5.0 can still import what it needs from Travertino."""
    with pytest.deprecated_call():
        from travertino.declaration import BaseStyle, Choices  # noqa


def test_apply_deprecated():
    """Calling with more than one argument is deprecated."""
    style = Style(explicit_const=VALUE2, implicit=VALUE3)
    style._applicator = Mock()
    style._apply.reset_mock()

    with pytest.warns(
        DeprecationWarning,
        match=(
            r"Calling Style\.apply\(\) with multiple arguments is deprecated\. Use the "
            r'"with Style\.batch_apply\(\):" context manager instead\.'
        ),
    ):
        style.apply("explicit_const", "implicit")

    # Should still call down to _apply, though.
    style._apply.assert_called_once_with({"explicit_const", "implicit"})



import pytest

from .style_classes import VALUE1, VALUE2, VALUE3, BaseStyle, DeprecatedStyle, Style


class StyleSubclass(Style):
    pass


class DeprecatedStyleSubclass(DeprecatedStyle):
    pass


class Sibling(BaseStyle):
    pass


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_dict(StyleClass):
    "Style declarations expose a dict-like interface"
    style = StyleClass()

    style.update(
        explicit_const=VALUE2,
        explicit_value=20,
        thing=(30, 40, 50, 60),
    )

    expected_keys = {
        "explicit_const",
        "explicit_value",
        "thing_bottom",
        "thing_left",
        "thing_right",
        "thing_top",
    }

    assert style.keys() == expected_keys

    assert sorted(style.items()) == sorted(
        [
            ("explicit_const", "value2"),
            ("explicit_value", 20),
            ("thing_bottom", 50),
            ("thing_left", 60),
            ("thing_right", 40),
            ("thing_top", 30),
        ]
    )

    # Style object has a length and is iterable.
    assert len(style) == 6
    for name in style:
        assert name in expected_keys

    # Properties that are set are in the keys.
    for name in expected_keys:
        assert name in style

    # Directional properties with one or more of the aliased properties set also count.
    assert "thing" in style

    # Valid properties that haven't been set are not in the keys.
    assert "implicit" not in style
    assert "explicit_none" not in style

    # Neither are invalid properties.
    assert "invalid_property" not in style

    # A property can be set, retrieved and cleared using the attribute name
    style["thing-bottom"] = 10
    assert style["thing-bottom"] == 10
    del style["thing-bottom"]
    assert style["thing-bottom"] == 0

    # A property can be set, retrieved and cleared using the Python attribute name
    style["thing_bottom"] = 10
    assert style["thing_bottom"] == 10
    del style["thing_bottom"]
    assert style["thing_bottom"] == 0

    # Property aliases can be accessed as well.
    style["thing"] = 5
    assert style["thing"] == (5, 5, 5, 5)
    del style["thing"]
    assert style["thing"] == (0, 0, 0, 0)

    # Clearing a valid property isn't an error
    del style["thing_bottom"]
    assert style["thing_bottom"] == 0

    # Non-existent properties raise KeyError
    with pytest.raises(KeyError):
        style["no-such-property"] = "no-such-value"

    with pytest.raises(KeyError):
        style["no-such-property"]

    with pytest.raises(KeyError):
        del style["no-such-property"]


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
@pytest.mark.parametrize("instantiate", [True, False])
def test_union_operators(StyleClass, instantiate):
    """Styles support | and |= with dicts and with their own class."""
    left = StyleClass(explicit_value=VALUE1, implicit=VALUE2)

    style_dict = {"thing_top": 5, "implicit": VALUE3}
    right = StyleClass(**style_dict) if instantiate else style_dict

    # Standard operator
    result = left | right

    # Original objects unchanged
    assert left["explicit_value"] == VALUE1
    assert left["implicit"] == VALUE2

    assert right["thing_top"] == 5
    assert right["implicit"] == VALUE3

    # Unshared properties assigned
    assert result["explicit_const"] == VALUE1
    assert result["thing_top"] == 5

    # Common property overridden by second operand
    assert result["implicit"] == VALUE3

    # In-place version
    left |= right

    # Common property updated on lefthand
    assert left["explicit_value"] == VALUE1
    assert left["implicit"] == VALUE3

    # Righthand unchanged
    assert right["thing_top"] == 5
    assert right["implicit"] == VALUE3


@pytest.mark.parametrize(
    "StyleClass, OtherClass",
    [
        (Style, StyleSubclass),
        (Style, Sibling),
        (Style, int),
        (Style, list),
        (DeprecatedStyle, DeprecatedStyleSubclass),
        (DeprecatedStyle, Sibling),
        (DeprecatedStyle, int),
        (DeprecatedStyle, list),
    ],
)
def test_union_operators_invalid_type(StyleClass, OtherClass):
    """Styles do not support | or |= with other style classes or with non-mappings."""

    left = StyleClass()
    right = OtherClass()

    with pytest.raises(TypeError, match=r"unsupported operand type"):
        left | right

    with pytest.raises(TypeError, match=r"unsupported operand type"):
        left |= right


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
@pytest.mark.parametrize(
    "right, error",
    [
        ({"implicit": "bogus_value"}, ValueError),
        ({"bogus_key": 3.12}, NameError),
    ],
)
def test_union_operators_invalid_key_value(StyleClass, right, error):
    """Operators will accept any mapping, but invalid keys/values are still an error."""
    left = StyleClass()

    with pytest.raises(error):
        left | right

    with pytest.raises(error):
        left |= right



from unittest.mock import call

import pytest

from .style_classes import VALUE1, VALUE2, VALUE3, DeprecatedStyle, Style


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_apply(StyleClass):
    style = StyleClass(explicit_const=VALUE2, implicit=VALUE3)
    assert style.apply.mock_calls == [call("explicit_const"), call("implicit")]
    style.apply.reset_mock()

    style.apply()
    style.apply.assert_called_once_with()


@pytest.mark.parametrize(
    "name, initial",
    [
        ("explicit_const", VALUE1),
        ("explicit_value", 0),
        ("explicit_none", None),
        ("implicit", None),
    ],
)
@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_property(name, initial, StyleClass):
    style = StyleClass()

    # Default value
    assert style[name] == initial
    style.apply.assert_not_called()

    # Modify the value
    style[name] = 10

    assert style[name] == 10
    style.apply.assert_called_once_with(name)

    # Clear the applicator mock
    style.apply.reset_mock()

    # Set the value to the same value.
    # No dirty notification is sent
    style[name] = 10
    assert style[name] == 10
    style.apply.assert_not_called()

    # Set the value to something new
    # A dirty notification is set.
    style[name] = 20
    assert style[name] == 20
    style.apply.assert_called_once_with(name)

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property
    del style[name]
    assert style[name] == initial
    style.apply.assert_called_once_with(name)

    # Clear the applicator mock
    style.apply.reset_mock()

    # Clear the property again.
    # The underlying attribute won't exist, so this should be a no-op.
    del style[name]
    assert style[name] is initial
    style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_set_delete_initial_no_apply(StyleClass):
    """Assigning or deleting a value equal to the initial value shouldn't apply."""
    style = StyleClass()

    # 0 is the initial value
    style.explicit_value = 0
    style.apply.assert_not_called()

    del style.explicit_value
    style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_set_multiple_properties(StyleClass):
    style = StyleClass()

    # Set a pair of properties
    style.update(explicit_value=20, explicit_none=10)

    assert style.explicit_const is VALUE1
    assert style.explicit_none == 10
    assert style.explicit_value == 20
    style.apply.assert_has_calls(
        [
            call("explicit_value"),
            call("explicit_none"),
        ],
        any_order=True,
    )

    # Set a different pair of properties
    style.update(explicit_const=VALUE2, explicit_value=30)

    assert style.explicit_const is VALUE2
    assert style.explicit_value == 30
    assert style.explicit_none == 10
    style.apply.assert_has_calls(
        [
            call("explicit_const"),
            call("explicit_value"),
        ],
        any_order=True,
    )

    # Clear the applicator mock
    style.apply.reset_mock()

    # Setting a non-property
    with pytest.raises(NameError):
        style.update(not_a_property=10)

    style.apply.assert_not_called()


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_set_to_initial(StyleClass):
    """A property set to its initial value is distinct from an unset property."""
    style = StyleClass()

    # explicit_const's initial value is VALUE1.
    assert style.explicit_const == VALUE1
    assert "explicit_const" not in style

    # The unset property shouldn't affect the value when overlaid over a style with
    # that property set.
    non_initial_style = StyleClass(explicit_const=VALUE2)
    union = non_initial_style | style
    assert union.explicit_const == VALUE2
    assert "explicit_const" in union

    # The property should count as set, even when set to the same initial value.
    style.explicit_const = VALUE1
    assert style.explicit_const == VALUE1
    assert "explicit_const" in style

    # The property should now overwrite.
    union = non_initial_style | style
    assert union.explicit_const == VALUE1
    assert "explicit_const" in union



import re
from contextlib import nullcontext
from functools import partial

import pytest

from .style_classes import VALUE1, VALUE2, VALUE3, Style


@pytest.mark.parametrize(
    "source, alias, initial, value, context",
    [
        (
            "explicit_const",
            "plain_alias",
            VALUE1,
            VALUE3,
            nullcontext,
        ),
        (
            "explicit_const",
            "plain_alias_deprecated",
            VALUE1,
            VALUE3,
            partial(
                pytest.deprecated_call,
                match=(
                    r"Style\.plain_alias_deprecated is deprecated\. "
                    r"Use Style\.explicit_const instead\."
                ),
            ),
        ),
        # There *shouldn't* be anything special about an alias to a
        # directional_property vs. to a validated_property, but it doesn't hurt to
        # check.
        (
            "thing",
            "directional_alias",
            (0, 0, 0, 0),
            (10, 5, 10, 5),
            nullcontext,
        ),
        (
            "thing",
            "directional_alias_deprecated",
            (0, 0, 0, 0),
            (10, 5, 10, 5),
            partial(
                pytest.deprecated_call,
                match=(
                    r"Style\.directional_alias_deprecated is deprecated\. "
                    r"Use Style\.thing instead\."
                ),
            ),
        ),
    ],
)
def test_simple_alias(source, alias, initial, value, context):
    """An alias with a constant source refers consistently to the source property."""
    style = Style()

    # Set on original, check on alias
    with context():
        assert alias not in style

    style[source] = value

    with context():
        assert alias in style
    with context():
        assert style[alias] == value

    del style[source]

    with context():
        assert alias not in style
    with context():
        assert style[alias] == initial

    # Set on alias, check on original.
    assert source not in style

    with context():
        style[alias] = value

    assert source in style
    assert style[source] == value

    with context():
        del style[alias]

    assert source not in style
    assert style[source] == initial


@pytest.mark.parametrize(
    "alias, context",
    [
        ("conditional_alias", nullcontext),
        ("conditional_alias_deprecated", pytest.deprecated_call),
    ],
)
@pytest.mark.parametrize(
    "properties, source, value",
    [
        ({}, "explicit_const", VALUE1),
        ({"thing_top": 10, "list_prop": [VALUE1, VALUE2]}, "explicit_value", 0),
        ({"thing_top": 10, "list_prop": [VALUE1]}, "explicit_none", None),
    ],
)
def test_conditional_alias(alias, context, properties, source, value):
    """An alias correctly checks conditions to determine which property to access."""
    style = Style(**properties)
    with context():
        assert style[alias] == style[source] == value


@pytest.mark.parametrize(
    "alias",
    # No deprecation warnings are given; there's no way to even tell which newer name
    # should be instead.
    ["conditional_alias", "conditional_alias_deprecated"],
)
@pytest.mark.parametrize(
    "properties",
    [
        {"thing_top": 1},
        {"thing_top": 10, "list_prop": [VALUE1, VALUE3]},
        {"thing_top": 11},
    ],
)
def test_conditional_alias_invald(alias, properties):
    """If no condition is valid, using the alias raises an AttributeError."""
    error_msg = re.escape(
        f"'{alias}' is only supported when (thing_top == 0) or (thing_top == 10; "
        "list_prop == ['value1', 'value2']) or (thing_top == 10; list_prop == "
        "['value1'])"
    )
    style = Style(**properties)
    with pytest.raises(AttributeError, match=error_msg):
        _ = style[alias]

    with pytest.raises(AttributeError, match=error_msg):
        style[alias] = 15

    with pytest.raises(AttributeError, match=error_msg):
        _ = alias in style

    with pytest.raises(AttributeError, match=error_msg):
        del style[alias]


@pytest.mark.parametrize(
    "alias, context",
    [
        ("conditional_alias", nullcontext),
        ("conditional_alias_deprecated", pytest.deprecated_call),
    ],
)
@pytest.mark.parametrize(
    "properties",
    [
        {"alias": 1, "thing_top": 10, "list_prop": [VALUE1]},
        {"thing_top": 10, "alias": 1, "list_prop": [VALUE1]},
        {"thing_top": 10, "list_prop": [VALUE1], "alias": 1},
    ],
)
def test_conditional_alias_simultaneous_setting(alias, context, properties):
    """Setting an alias along with its conditions works, in any order."""
    properties = {
        (alias if name == "alias" else name): value
        for name, value in properties.items()
    }
    with context():
        style = Style(**properties)
    with context():
        assert style[alias] == 1

    style = Style()
    with context():
        style.update(**properties)
    with context():
        assert style[alias] == 1



import pytest

from travertino.properties.validated import list_property, validated_property

from .style_classes import (
    VALUE2,
    VALUE3,
    VALUES,
    DeprecatedStyle,
    Style,
)


def test_invalid_style():
    with pytest.raises(ValueError):
        # Define an invalid initial value on a validated property
        validated_property(*VALUES, integer=True, initial="something")

    with pytest.raises(ValueError):
        # Same for list property
        list_property(*VALUES, integer=True, initial=["something"])


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_positional_argument(StyleClass):
    # Could be the subclass or inherited __init__, depending on Python version / API
    # used.
    with pytest.raises(
        TypeError, match=r"__init__\(\) takes 1 positional argument but 2 were given"
    ):
        StyleClass(5)


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_constructor_invalid_property(StyleClass):
    """Whether dataclass or not, the error should be the same."""
    with pytest.raises(
        TypeError,
        match=r"Style\.__init__\(\) got an unexpected keyword argument 'bogus'",
    ):
        StyleClass(explicit_const=5, bogus=None)


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_create_and_copy(StyleClass):
    style = StyleClass(explicit_const=VALUE2, implicit=VALUE3)

    dup = style.copy()
    assert dup.explicit_const == VALUE2
    assert dup.explicit_value == 0
    assert dup.implicit == VALUE3


@pytest.mark.parametrize("StyleClass", [Style, DeprecatedStyle])
def test_str(StyleClass):
    style = StyleClass()

    style.update(
        explicit_const=VALUE2,
        explicit_value=20,
        thing=(30, 40, 50, 60),
    )

    assert (
        str(style) == "explicit-const: value2; "
        "explicit-value: 20; "
        "thing-bottom: 50; "
        "thing-left: 60; "
        "thing-right: 40; "
        "thing-top: 30"
    )


def test_repr():
    # Doesn't need to be tested with deprecated API.
    style = Style(explicit_const=VALUE2, explicit_value=20, thing=(30, 40, 50, 60))

    assert repr(style) == (
        "Style("
        "explicit_const='value2', "
        "explicit_value=20, "
        "thing_bottom=50, "
        "thing_left=60, "
        "thing_right=40, "
        "thing_top=30"
        ")"
    )



from collections.abc import Sequence

import pytest

from travertino.properties.immutablelist import ImmutableList

from .style_classes import VALUE1, VALUE2, VALUE3, Style


@pytest.mark.parametrize(
    "value, expected",
    [
        ([VALUE1], [VALUE1]),
        (VALUE1, [VALUE1]),
        ([VALUE1, VALUE3], [VALUE1, VALUE3]),
        ([VALUE2, VALUE1], [VALUE2, VALUE1]),
        ([VALUE2, VALUE3, 1, 2, VALUE1], [VALUE2, VALUE3, 1, 2, VALUE1]),
        # Duplicates are kept, but "normalized" via validation.
        (
            [VALUE3, 1, VALUE3, "1", True, " 1", VALUE2],
            [VALUE3, 1, VALUE3, 1, 1, 1, VALUE2],
        ),
        # Other sequences should work too.
        ((VALUE1, VALUE3), [VALUE1, VALUE3]),
    ],
)
def test_list_property(value, expected):
    style = Style()
    style.list_prop = value
    assert style.list_prop == expected


@pytest.mark.parametrize(
    "value, error, match",
    [
        (
            5,
            TypeError,
            r"Value for list property list_prop must be a sequence\.",
        ),
        (
            # Fails because it's only a generator, not a comprehension:
            (i for i in [VALUE1, VALUE3]),
            TypeError,
            r"Value for list property list_prop must be a sequence.",
        ),
        (
            [VALUE3, VALUE1, "bogus"],
            ValueError,
            r"Invalid item value 'bogus' for list property list_prop; "
            r"Valid values are: none, value1, value2, value3, <integer>",
        ),
        (
            (),
            ValueError,
            r"List properties cannot be set to an empty sequence; "
            r"to reset a property, use del `style.list_prop`\.",
        ),
        (
            [],
            ValueError,
            r"List properties cannot be set to an empty sequence; "
            r"to reset a property, use del `style.list_prop`\.",
        ),
    ],
)
def test_list_property_invalid(value, error, match):
    style = Style()
    with pytest.raises(error, match=match):
        style.list_prop = value


def test_list_property_immutable():
    style = Style()
    style.list_prop = [1, 2, 3, VALUE2]
    prop = style.list_prop

    with pytest.raises(TypeError, match=r"does not support item assignment"):
        prop[0] = 5

    with pytest.raises(TypeError, match=r"doesn't support item deletion"):
        del prop[1]

    with pytest.raises(AttributeError):
        prop.insert(2, VALUE1)

    with pytest.raises(AttributeError):
        prop.append(VALUE3)

    with pytest.raises(AttributeError):
        prop.clear()

    with pytest.raises(AttributeError):
        prop.reverse()

    with pytest.raises(AttributeError):
        prop.pop()

    with pytest.raises(AttributeError):
        prop.remove(VALUE2)

    with pytest.raises(AttributeError):
        prop.extend([5, 6, 7])

    with pytest.raises(TypeError, match=r"unsupported operand type\(s\)"):
        prop += [4, 3, VALUE1]

    with pytest.raises(TypeError, match=r"unsupported operand type\(s\)"):
        prop += ImmutableList([4, 3, VALUE1])

    with pytest.raises(AttributeError):
        prop.sort()


def test_list_property_list_like():
    style = Style()
    style.list_prop = [1, 2, 3, VALUE2]
    prop = style.list_prop

    assert isinstance(prop, ImmutableList)
    assert prop == [1, 2, 3, VALUE2]
    assert prop == ImmutableList([1, 2, 3, VALUE2])
    assert prop[2] == 3
    assert str(prop) == repr(prop) == "[1, 2, 3, 'value2']"
    assert len(prop) == 4

    count = 0
    for _ in prop:
        count += 1
    assert count == 4

    assert [*reversed(prop)] == [VALUE2, 3, 2, 1]

    assert prop.index(3) == 2

    assert prop.count(VALUE2) == 1

    assert isinstance(prop, Sequence)
